# frozen_string_literal: true

require 'numo/narray'

require_relative 'rumale/version'
require_relative 'rumale/rumaleext'
require_relative 'rumale/validation'
require_relative 'rumale/values'
require_relative 'rumale/utils'
require_relative 'rumale/pairwise_metric'
require_relative 'rumale/dataset'
require_relative 'rumale/probabilistic_output'
require_relative 'rumale/base/base_estimator'
require_relative 'rumale/base/classifier'
require_relative 'rumale/base/regressor'
require_relative 'rumale/base/cluster_analyzer'
require_relative 'rumale/base/transformer'
require_relative 'rumale/base/splitter'
require_relative 'rumale/base/evaluator'
require_relative 'rumale/pipeline/pipeline'
require_relative 'rumale/pipeline/feature_union'
require_relative 'rumale/kernel_approximation/rbf'
require_relative 'rumale/kernel_approximation/nystroem'
require_relative 'rumale/linear_model/base_sgd'
require_relative 'rumale/linear_model/svc'
require_relative 'rumale/linear_model/svr'
require_relative 'rumale/linear_model/logistic_regression'
require_relative 'rumale/linear_model/linear_regression'
require_relative 'rumale/linear_model/ridge'
require_relative 'rumale/linear_model/lasso'
require_relative 'rumale/linear_model/elastic_net'
require_relative 'rumale/linear_model/nnls'
require_relative 'rumale/kernel_machine/kernel_svc'
require_relative 'rumale/kernel_machine/kernel_pca'
require_relative 'rumale/kernel_machine/kernel_fda'
require_relative 'rumale/kernel_machine/kernel_ridge'
require_relative 'rumale/kernel_machine/kernel_ridge_classifier'
require_relative 'rumale/multiclass/one_vs_rest_classifier'
require_relative 'rumale/nearest_neighbors/vp_tree'
require_relative 'rumale/nearest_neighbors/k_neighbors_classifier'
require_relative 'rumale/nearest_neighbors/k_neighbors_regressor'
require_relative 'rumale/naive_bayes/base_naive_bayes'
require_relative 'rumale/naive_bayes/bernoulli_nb'
require_relative 'rumale/naive_bayes/complement_nb'
require_relative 'rumale/naive_bayes/gaussian_nb'
require_relative 'rumale/naive_bayes/multinomial_nb'
require_relative 'rumale/naive_bayes/negation_nb'
require_relative 'rumale/tree/node'
require_relative 'rumale/tree/base_decision_tree'
require_relative 'rumale/tree/decision_tree_classifier'
require_relative 'rumale/tree/decision_tree_regressor'
require_relative 'rumale/tree/extra_tree_classifier'
require_relative 'rumale/tree/extra_tree_regressor'
require_relative 'rumale/tree/gradient_tree_regressor'
require_relative 'rumale/ensemble/ada_boost_classifier'
require_relative 'rumale/ensemble/ada_boost_regressor'
require_relative 'rumale/ensemble/gradient_boosting_classifier'
require_relative 'rumale/ensemble/gradient_boosting_regressor'
require_relative 'rumale/ensemble/random_forest_classifier'
require_relative 'rumale/ensemble/random_forest_regressor'
require_relative 'rumale/ensemble/extra_trees_classifier'
require_relative 'rumale/ensemble/extra_trees_regressor'
require_relative 'rumale/ensemble/stacking_classifier'
require_relative 'rumale/ensemble/stacking_regressor'
require_relative 'rumale/ensemble/voting_classifier'
require_relative 'rumale/ensemble/voting_regressor'
require_relative 'rumale/clustering/k_means'
require_relative 'rumale/clustering/mini_batch_k_means'
require_relative 'rumale/clustering/k_medoids'
require_relative 'rumale/clustering/gaussian_mixture'
require_relative 'rumale/clustering/dbscan'
require_relative 'rumale/clustering/hdbscan'
require_relative 'rumale/clustering/snn'
require_relative 'rumale/clustering/power_iteration'
require_relative 'rumale/clustering/spectral_clustering'
require_relative 'rumale/clustering/single_linkage'
require_relative 'rumale/decomposition/pca'
require_relative 'rumale/decomposition/nmf'
require_relative 'rumale/decomposition/factor_analysis'
require_relative 'rumale/decomposition/fast_ica'
require_relative 'rumale/manifold/tsne'
require_relative 'rumale/manifold/mds'
require_relative 'rumale/metric_learning/fisher_discriminant_analysis'
require_relative 'rumale/metric_learning/neighbourhood_component_analysis'
require_relative 'rumale/metric_learning/mlkr'
require_relative 'rumale/neural_network/adam'
require_relative 'rumale/neural_network/base_mlp'
require_relative 'rumale/neural_network/mlp_regressor'
require_relative 'rumale/neural_network/mlp_classifier'
require_relative 'rumale/feature_extraction/hash_vectorizer'
require_relative 'rumale/feature_extraction/feature_hasher'
require_relative 'rumale/feature_extraction/tfidf_transformer'
require_relative 'rumale/preprocessing/l2_normalizer'
require_relative 'rumale/preprocessing/l1_normalizer'
require_relative 'rumale/preprocessing/max_normalizer'
require_relative 'rumale/preprocessing/min_max_scaler'
require_relative 'rumale/preprocessing/max_abs_scaler'
require_relative 'rumale/preprocessing/standard_scaler'
require_relative 'rumale/preprocessing/bin_discretizer'
require_relative 'rumale/preprocessing/label_binarizer'
require_relative 'rumale/preprocessing/label_encoder'
require_relative 'rumale/preprocessing/one_hot_encoder'
require_relative 'rumale/preprocessing/ordinal_encoder'
require_relative 'rumale/preprocessing/binarizer'
require_relative 'rumale/preprocessing/polynomial_features'
require_relative 'rumale/preprocessing/kernel_calculator'
require_relative 'rumale/model_selection/k_fold'
require_relative 'rumale/model_selection/group_k_fold'
require_relative 'rumale/model_selection/stratified_k_fold'
require_relative 'rumale/model_selection/shuffle_split'
require_relative 'rumale/model_selection/group_shuffle_split'
require_relative 'rumale/model_selection/stratified_shuffle_split'
require_relative 'rumale/model_selection/time_series_split'
require_relative 'rumale/model_selection/cross_validation'
require_relative 'rumale/model_selection/grid_search_cv'
require_relative 'rumale/model_selection/function'
require_relative 'rumale/evaluation_measure/accuracy'
require_relative 'rumale/evaluation_measure/precision'
require_relative 'rumale/evaluation_measure/recall'
require_relative 'rumale/evaluation_measure/f_score'
require_relative 'rumale/evaluation_measure/roc_auc'
require_relative 'rumale/evaluation_measure/log_loss'
require_relative 'rumale/evaluation_measure/r2_score'
require_relative 'rumale/evaluation_measure/explained_variance_score'
require_relative 'rumale/evaluation_measure/mean_squared_error'
require_relative 'rumale/evaluation_measure/mean_squared_log_error'
require_relative 'rumale/evaluation_measure/mean_absolute_error'
require_relative 'rumale/evaluation_measure/median_absolute_error'
require_relative 'rumale/evaluation_measure/adjusted_rand_score'
require_relative 'rumale/evaluation_measure/purity'
require_relative 'rumale/evaluation_measure/mutual_information'
require_relative 'rumale/evaluation_measure/normalized_mutual_information'
require_relative 'rumale/evaluation_measure/silhouette_score'
require_relative 'rumale/evaluation_measure/davies_bouldin_score'
require_relative 'rumale/evaluation_measure/calinski_harabasz_score'
require_relative 'rumale/evaluation_measure/function'
